#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""The Infinity value.

Not all values are comparable with `Infinity`:obj: by default.  The ABC
`InfinityComparable`:class: holds the registry of such values.  Any `number
<numbers.Number>`:class: is comparable with `Infinity`:obj:.

`Dates <datetime.date>`:class:, `datetimes <datetime.datetime>`:class: and
`time deltas <datetime.timedelta>`:class: are also registered by default.

.. warning:: In Python 2, dates, datetimes and time deltas must be the right
             operand, as in ``Infinity > today``.  Doing ``today < Infinity``
             fails in Python 2.

"""

import abc
import datetime
from numbers import Number
from functools import total_ordering


class InfinityComparable(metaclass=abc.ABCMeta):
    """Any type that can be sensibly compared to infinity.

    All types in the `number <numbers.Number>`:class: tower are *always*
    comparable.

    Classes `datetime.date`:class:, `datetime.datetime`:class:, and
    `datetime.timedelta`:class: are automatically registered.

    """

    @classmethod
    def __subclasshook__(self, cls):
        if isinstance(cls, type) and issubclass(cls, Number):
            return True
        else:
            return NotImplemented


InfinityComparable.register(datetime.date)
InfinityComparable.register(datetime.datetime)
InfinityComparable.register(datetime.timedelta)


@total_ordering
class InfinityType:
    _positive = None
    _negative = None

    def __new__(cls, sign):
        if sign < 0:
            res = cls._negative
            if not res:
                cls._negative = res = object.__new__(cls)
        else:
            res = cls._positive
            if not res:
                cls._positive = res = object.__new__(cls)
        return res

    def __init__(self, sign):
        self.sign = -1 if sign < 0 else 1

    def __lt__(self, other):
        if isinstance(other, InfinityComparable):
            return self.sign < 0  # True iff -Infinity
        elif isinstance(other, InfinityType):
            return self.sign < other.sign
        else:
            raise TypeError("Incomparable types: %r and %r" % (type(self), type(other)))

    def __eq__(self, other):
        if isinstance(other, InfinityType):
            return self.sign == other.sign
        else:
            return NotImplemented

    def __neg__(self):
        return type(self)(-self.sign)

    def __str__(self):
        return "∞" if self.sign > 0 else "-∞"

    def __repr__(self):
        return "Infinity" if self.sign > 0 else "-Infinity"

    def __hash__(self):
        return id(self)


Infinity = InfinityType(+1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Some useful decorators.'''

# TODO: reconsider all this module


import sys

from functools import wraps
from types import FunctionType as function

from xotl.tools.decorator.meta import decorator


__all__ = ('decorator', 'AttributeAlias', 'settle', 'namer', 'aliases',
           'assignment_operator', 'instantiate',
           'memoized_instancemethod', 'reset_memoized')


class AttributeAlias:
    '''Descriptor to create aliases for object attributes.

    This descriptor is mainly to be used internally by `aliases`:func:
    decorator.

    '''

    def __init__(self, attr_name):
        super().__init__()
        self.attr_name = attr_name

    def __get__(self, instance, owner):
        return getattr(instance or owner, self.attr_name)

    def __set__(self, instance, value):
        setattr(instance, self.attr_name, value)

    def __delete__(self, instance):
        delattr(instance, self.attr_name)


def settle(**kwargs):
    '''Returns a decorator to settle attributes to the decorated target.

    Usage::

       >>> @settle(name='Name')
       ... class Person:
       ...    pass

       >>> Person.name
       'Name'

    '''
    def inner(target):
        for attr in kwargs:
            setattr(target, attr, kwargs[attr])
        return target
    return inner


def namer(name, **kwargs):
    '''Like `settle`:func:, but '__name__' is a required positional argument.

    Usage::

        >>> @namer('Identity', custom=1)
        ... class I:
        ...    pass

        >>> I.__name__
        'Identity'

        >>> I.custom
        1

    '''
    return settle(__name__=name, **kwargs)


def aliases(*names, **kwargs):
    '''In a class, create an `AttributeAlias`:class: descriptor for each
    definition as keyword argument (alias=existing_attribute).

    If "names" are given, then the definition context is looked and are
    assigned to it the same decorator target with all new names::

        >>> @aliases('foo', 'bar')
        ... def foobar(*args):
        ...     'This function is added to its module with two new names.'

    '''
    # FIX: This is not working in methods.
    def inner(target):
        '''Direct closure decorator that settle several attribute aliases.'''
        if kwargs:
            assert isinstance(target, type), '"target" must be a class.'
        if names:
            _locals = sys._getframe(1).f_locals
            for name in names:
                _locals[str(name)] = target
        if kwargs:
            for alias in kwargs:
                field = kwargs[alias]
                setattr(target, alias, AttributeAlias(field))
        return target
    return inner


@decorator
def assignment_operator(func, maybe_inline=False):
    '''Makes a function that receives a name, and other args to get its first
    argument (the name) from an assignment operation, meaning that it if its
    used in a single assignment statement the name will be taken from the left
    part of the ``=`` operator.

    .. warning:: This function is dependant of CPython's implementation of the
                 language and won't probably work on other implementations.
                 Use only you don't care about portability, but use sparingly
                 (in case you change your mind about portability).

    '''
    import inspect
    import ast

    if not isinstance(func, function):
        raise TypeError('"func" must be a function.')

    @wraps(func)
    def inner(*args):
        frm = sys._getframe(1)
        filename, line, funcname, src_lines, idx = inspect.getframeinfo(frm)
        try:
            sourceline = src_lines[idx].strip()
            parsed_line = ast.parse(sourceline, filename).body[0]
            assert maybe_inline or isinstance(parsed_line, ast.Assign)
            if isinstance(parsed_line, ast.Assign):
                assert len(parsed_line.targets) == 1
                assert isinstance(parsed_line.targets[0], ast.Name)
                name = parsed_line.targets[0].id
            elif maybe_inline:
                assert isinstance(parsed_line, ast.Expr)
                name = None
            else:
                assert False
            return func(name, *args)
        except (AssertionError, SyntaxError):
            if maybe_inline:
                return func(None, *args)
            else:
                return func(*args)
        finally:
            del filename, line, funcname, src_lines, idx
    return inner


@decorator
def instantiate(target, *args, **kwargs):
    '''Some singleton classes must be instantiated as part of its declaration
    because they represents singleton objects.

    Every argument, positional or keyword, is passed as such when invoking the
    target. The following two code samples show two cases::

       >>> @instantiate
       ... class Foobar:
       ...    def __init__(self):
       ...        print('Init...')
       Init...


       >>> @instantiate('test', context={'x': 1})
       ... class Foobar:
       ...    def __init__(self, name, context):
       ...        print('Initializing a Foobar instance with name={name!r} '
       ...              'and context={context!r}'.format(**locals()))
       Initializing a Foobar instance with name='test' and context={'x': 1}

    In all cases, Foobar remains the class, not the instance::

        >>> Foobar  # doctest: +ELLIPSIS
        <class '...Foobar'>

    '''
    target(*args, **kwargs)
    return target


@decorator
def constant_bagger(func, *args, **kwds):
    '''Create a "bag" with constant values.

    Decorated object must be a callable, but the result will be a class
    containing the constant values.

    For example::

      >>> @constant_bagger
      ... def MYBAG():
      ...     return dict(X=1, Y=2)

    It will generate::

      class MYBAG:
          X = 1
          Y = 2

    When called with arguments, these will be used as actual arguments for the
    decorated function::

      >>> @constant_bagger(X=1, Y=2)
      ... def MYBAG(**kwds):
      ...     return kwds

    Constant bags are singletons that can be updated::

      >>> MYBAG(Z=3) is MYBAG
      True

      >>> MYBAG.Z
      3

    '''
    from xotl.tools.objects import mass_setattr
    wraps = ((a, getattr(func, a, None)) for a in ('__doc__', '__module__'))
    attrs = {a: v for (a, v) in wraps if v}
    attrs.update(__new__=mass_setattr, **func(*args, **kwds))
    return type(func.__name__, (object,), attrs)


@decorator
def singleton(target, *args, **kwargs):
    '''Instantiate a class and assign the instance to the declared symbol.

    Every argument, positional or keyword, is passed as such when invoking the
    target. The following two code samples show two cases::

      >>> @singleton
      ... class foobar:
      ...    def __init__(self):
      ...        self.doc = 'foobar instance'

      >>> foobar.doc
      'foobar instance'

      >>> @singleton('test', context={'x': 1})
      ... class foobar:
      ...     def __init__(self, name, context):
      ...         self.name = name
      ...         self.context = context

      >>> foobar.name, foobar.context
      ('test', {'x': 1})

      >>> isinstance(foobar, type)
      False

    '''
    res = target(*args, **kwargs)
    if isinstance(target, type):
        try:
            def __init__(*args, **kwds):
                msg = "'{}' is a singleton, it can be instantiated only once"
                raise TypeError(msg.format(target.__name__))
            target.__init__ = __init__
        except Exception:
            pass
    return res


class memoized_instancemethod:
    """Decorate a method memoize its return value.

    Best applied to no-arg methods: memoization is not sensitive to
    argument values, and will always return the same value even when
    called with different arguments.

    This is extracted from the SQLAlchemy project's codebase, merit and
    copyright goes to SQLAlchemy authors::

      Copyright (C) 2005-2011 the SQLAlchemy authors and contributors

      This module is part of SQLAlchemy and is released under the MIT License:
      http://www.opensource.org/licenses/mit-license.php

    """
    def __init__(self, fget, doc=None):
        from warnings import warn
        msg = ('"memoized_instancemethod" is now deprecated and it will be '
               'removed.')
        warn(msg, stacklevel=2)
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self

        def oneshot(*args, **kw):
            result = self.fget(obj, *args, **kw)
            memo = lambda *a, **kw: result
            memo.__name__ = self.__name__
            memo.__doc__ = self.__doc__
            obj.__dict__[self.__name__] = memo
            return result

        oneshot.__name__ = self.__name__
        oneshot.__doc__ = self.__doc__
        return oneshot


def reset_memoized(instance, name):
    from warnings import warn
    msg = ('"reset_memoized" is now deprecated and it will be '
           'removed. Use "memoized_property.reset".')
    warn(msg, stacklevel=2)
    instance.__dict__.pop(name, None)


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Validity proofs for data values.

There are some basic helper functions:

- `predicative`:func: wraps a function in a way that a logical false value is
  returned on failure.  If an exception is raised, it is returned wrapped as
  an special false value.  See `~xotl.tools.fp.option.Maybe`:class: monad for
  more information.

- `vouch`:func: wraps a function in a way that an exception is raised if
  an invalid value (logical false by default) is returned.  This is useful to
  call functions that use "special" false values to signal a failure.

- `enfold`:func: creates a decorator to convert a function to use either the
  `predicative`:func: or the `vouch`:func: protocol.

.. versionadded:: 1.8.0

'''


def predicative(function, *args, **kwds):
    '''Call a function in a safety wrapper returning a false value if fail.

    This converts any function into a predicate.  A predicate can be thought
    as an operator or function that returns a value that is either true or
    false.

    Predicates are sometimes used to indicate set membership: on certain
    occasions it is inconvenient or impossible to describe a set by listing
    all of its elements.  Thus, a predicate ``P(x)`` will be true or false,
    depending on whether x belongs to a set.

    If the argument `function` validates its arguments, return a valid true
    value.  There are two special conditions: first, a value treated as false
    for Python conventions (for example, ``0``, or an empty string); and
    second, when an exception is raised; in both cases the predicate will
    return an instance of `~xotl.tools.fp.option.Maybe`:class:.

    '''
    from xotl.tools.symbols import boolean
    from xotl.tools.fp.option import Maybe, Just, Wrong
    from xotl.tools.params import single
    # I don't understand anymore why a single argument must be a special case,
    # maybe because the composition problem.
    is_single = single(*args, **kwds)
    try:
        res = function(*args, **kwds)
        if isinstance(res, (boolean, Maybe)):
            if isinstance(res, Just) and res.inner:
                return res.inner
            elif isinstance(res, boolean) and is_single and args[0]:
                return args
            else:
                return res
        elif res:
            return res
        else:
            return Just(res)
    except Exception as error:
        if isinstance(error, ValueError) and is_single:
            return Wrong(args[0])
        else:
            return Wrong(error)


def vouch(function, *args, **kwds):
    '''Call a function in a safety wrapper raising an exception if it fails.

    When the wrapped function fails, an exception must be raised.  A predicate
    fails when it returns a false value.  To avoid treat false values of some
    types as fails, use `Just`:class: to return that values wrapped.

    '''
    from xotl.tools.symbols import boolean, Invalid
    from xotl.tools.clipping import small
    from xotl.tools.fp.option import Just, Wrong
    from xotl.tools.params import single
    res = function(*args, **kwds)
    if isinstance(res, boolean):
        if res:
            aux = single(*args, **kwds)
            if aux is not Invalid:
                res = aux
        else:
            msg = '{}() validates as false'.format(small(function))
            raise TypeError(msg)
    elif isinstance(res, Wrong):
        inner = res.inner
        if isinstance(inner, BaseException):
            raise inner
        else:
            msg = '{}() validates as a wrong value'.format(small(function))
            if inner is not None or not isinstance(inner, boolean):
                v, t = small(inner), type(inner).__name__
                msg += ' {} of type "{}"'.format(v, t)
            raise TypeError(msg)
    elif isinstance(res, Just):
        res = res.inner
    return res


def enfold(checker):
    '''Create a decorator to execute a function inner a safety wrapper.

    :param checker: Could be any function to enfold, but it's intended mainly
           for `predicative`:func:  or `vouch`:func: functions.

    In the following example, the semantics of this function can be seen.  The
    definition::

        >>> @enfold(predicative)
        ... def test(x):
        ...     return 1 <= x <= 10

        >>> test(5)
        5

    It is equivalent to::

        >>> def test(x):
        ...     return 1 <= x <= 10

        >>> predicative(test, 5)
        5

    In other hand::

        >>> @enfold(predicative)
        ... def test(x):
        ...     return 1 <= x <= 10

        >>> test(15)
        5

    '''
    def wrapper(func):
        def inner(*args, **kwds):
            return checker(func, *args, **kwds)

        try:
            inner.__name__ = func.__name__
            inner.__doc__ = func.__doc__
        except Exception:
            from xotl.tools.clipping import small
            inner.__name__ = str(small(func))
        return inner
    return wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Python context with thread-safe queued data.

'''


# TODO: Optimize this by using standard threading locks
class SafeData:
    '''Python context with queued data.'''
    __slots__ = ('queue', 'timeout', 'data',)

    def __init__(self, data, timeout=None):
        from queue import Queue
        self.queue = Queue(1)
        self.queue.put(data)
        self.timeout = timeout
        self.data = data

    def __enter__(self):
        res = self.queue.get(True, self.timeout)
        if res is self.data:
            return res
        else:
            raise RuntimeError('unexpected error, invalid queued data')

    def __exit__(self, exc_type, exc_value, traceback):
        data = self.data
        self.queue.task_done()
        self.queue.put(data, True, self.timeout)
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Some generic coercers (or checkers) for value types.

This module coercion function are not related in any way to deprecated old
python feature, are similar to a combination of object mold/check:

- *Mold* - Fit values to expected conventions.

- *Check* - These functions must return `nil` [#pyni]_ special value to
  specify that expected fit is not possible.

.. [#pyni] We don't use Python classic `NotImplemented` special value in order
           to obtain False if the value is not coerced (`nil`).

A custom coercer could be created with closures, for an example see
`create_int_range_coerce`:func:.

This module uses `Unset` value to define absent -not being specified-
arguments.

Also contains sub-modules to obtain, convert and check values of common types.

.. versionadded:: 1.7.0

'''

import re
from abc import ABCMeta

from xotl.tools.future.functools import lwraps
from xotl.tools.symbols import boolean, Unset
from xotl.tools.fp.prove import vouch

from xotl.tools.deprecation import deprecate_linked
deprecate_linked(check='xotl.tools.values')
del deprecate_linked


_coercer_decorator = lwraps(__coercer__=True)    # FIX: refactor


class logical(boolean):
    '''Represent Common Lisp two special values `t` and `nil`.

    Include redefinition of `__call__`:meth: to check values with special
    semantic:

    - When called as ``t(arg)``, check if `arg` is not `nil` returning a
      logical true: the same argument if `arg` is nil or a true boolean value,
      else return `t`.  That means that `False` or `0` are valid true values
      for Common Lisp but not for Python.

    - When called as ``nil(arg)``, check if `arg` is `nil` returning `t` or
      `nil` if not.

    Constructor could receive a valid name ('nil' or 't') or any other
    ``boolean`` instance.

    '''
    __slots__ = ()
    _valid = {'nil': False, 't': True}

    def __new__(cls, arg):
        from xotl.tools.symbols import boolean
        from xotl.tools.symbols import Invalid
        name = ('t' if arg else 'nil') if isinstance(arg, boolean) else arg
        value = cls._valid.get(name, Invalid)
        if value is not Invalid:
            return super().__new__(cls, name, value)
        else:
            msg = 'retrieving invalid logical instance "{}"'
            raise TypeError(msg.format(arg))

    def __call__(self, arg):
        if self:    # self is t
            return arg if arg or arg is nil else self
        else:    # self is nil
            return t if arg is self else self


nil, t = logical('nil'), logical('t')


class MetaCoercer(ABCMeta):
    r'''Meta-class for `coercer`:class:.

    This meta-class allows that several objects are considered valid instances
    of `coercer`:class:\ :

    - Functions decorated with `coercer`:class: (used with its decorator
      facet).

    - Instances of any sub-class of `custom`:class:.

    - Instances of `coercer`:class: itself.

    See the class declaration (`coercer`:class:) for more information.

    '''
    def __instancecheck__(self, instance):
        return (getattr(instance, '__coercer__', False) or
                super().__instancecheck__(instance))


class coercer(metaclass=MetaCoercer):
    '''Special coercer class.

    This class has several facets:

    - Pure type-checkers when a type or tuple of types are received as
      argument.  See `istype`:class: for more information.

    - Return equivalent coercer from some special values:

      * Any true value -> identity_coerce

      * Any false or empty value -> void_coerce

    - A decorator for functions; when a function is given, decorate it to
      become a coercer.  The mark itself is not enough, functions intended to
      be coercers must fulfills the protocol (not to produce exception and
      return `nil` on fails).  For example::

        >>> @coercer
        ... def age_coerce(arg):
        ...     res = int_coerce(arg)
        ...     return res if t(res) and 0 < arg <= 120 else nil

        # TODO: Change next, don't use isinstance
        >>> isinstance(age_coerce, coercer)
        True

    '''
    __slots__ = ()
    __coercer__ = True

    def __new__(cls, source):
        from types import FunctionType as function
        from xotl.tools.symbols import boolean
        if source == 1 and isinstance(source, boolean):
            return identity_coerce
        elif source is None or (source == 0 and isinstance(source, boolean)):
            return void_coerce
        elif isinstance(source, coercer):    # TODO: don't use isinstance
            return source
        elif isinstance(source, (function, staticmethod, classmethod)):
            return _coercer_decorator(source)
        else:
            inner = types_tuple_coerce(source)
            return istype(inner) if inner else nil


def coercer_name(arg, join=None):
    '''Get the name of a coercer.

    :param arg: Coercer to get the name.  Also processes collections (tuple,
           list, or set) of coercers.  Any other value is considered invalid
           and raises an exception.

    :param join: When a collection is used; if this argument is None a
           collection of names is returned, if not None then is used to join
           the items in a resulting string.

           For example::

             >>> coercer_name((int_coerce, float_coerce))
             ('int', 'float')

             >>> coercer_name((int_coerce, float_coerce), join='-')
             'int-float'

           To obtain pretty-print tuples, use something like::

             >>> coercer_name((int_coerce, float_coerce),
             ...              join=lambda arg: '(%s)' % ', '.join(arg))

    This function not only works with coercers, all objects that fulfill
    needed protocol to get names will also be valid.

    '''
    # TODO: Maybe this function must be moved to `xotl.tools.names`
    if isinstance(arg, (tuple, list, set)):
        res = type(arg)(coercer_name(c) for c in arg)
        if isinstance(join, str):
            join = join.join
        return str(join(res)) if join else res
    else:
        try:
            res = arg.__name__
        except Exception:
            res = str(arg)
        suffix = str('_coerce')
        if res.endswith(suffix):
            res = res[:-len(suffix)]
        return res


@coercer
def identity_coerce(arg):
    'Leaves unchanged the passed argument `arg`.'
    return arg


@coercer
def void_coerce(arg):
    '''Always `nil`.'''
    return nil


@coercer
def type_coerce(arg):
    '''Check if `arg` is a valid type.'''
    return arg if isinstance(arg, type) else nil


@coercer
def types_tuple_coerce(arg):
    '''Check if `arg` is valid for `isinstance` or `issubclass` 2nd argument.

    Type checkers are any class, a type or tuple of types.  For example::

      >>> types_tuple_coerce(object) == (object,)
      True

      >>> types_tuple_coerce((int, float)) == (int, float)
      true

      >>> types_tuple_coerce('not-a-type') is nil
      True

    See `type_coerce` for more information.

    '''
    if t(type_coerce(arg)):
        return (arg, )
    elif isinstance(arg, tuple) and all(t(type_coerce(tp)) for tp in arg):
        return arg
    else:
        return nil


@coercer
def callable_coerce(arg):
    '''Check if `arg` is a callable object.'''
    return arg if callable(arg) else nil


@coercer
def file_coerce(arg):
    '''Check if `arg` is a file-like object.'''
    from io import IOBase
    METHODS = ('close', 'write', 'read')
    ok = isinstance(arg, IOBase) or all(hasattr(arg, a) for a in METHODS)
    return arg if ok else nil


@coercer
def float_coerce(arg):
    '''Check if `arg` is a valid float.

    Other types are checked (string, int, complex).

    '''
    if isinstance(arg, float):
        return arg
    elif isinstance(arg, int):
        return float(arg)
    elif isinstance(arg, (str, bytes)):
        try:
            return float(arg)
        except ValueError:
            return nil
    elif isinstance(arg, complex):
        return arg.real if arg.imag == 0 else nil
    else:
        return nil


@coercer
def int_coerce(arg):
    '''Check if `arg` is a valid integer.

    Other types are checked (string, float, complex).

    '''
    if isinstance(arg, int):
        return arg
    else:
        arg = float_coerce(arg)
        if t(arg):
            res = int(arg)
            return res if arg - res == 0 else nil
        else:
            return nil


@coercer
def number_coerce(arg):
    '''Check if `arg` is a valid number (integer or float).

    Types that are checked (string, int, float, complex).

    '''
    if isinstance(arg, int):
        return arg
    else:
        f = float_coerce(arg)
        if t(f):
            i = int(f)
            return i if f - i == 0 else f
        else:
            return nil


@coercer
def positive_int_coerce(arg):
    '''Check if `arg` is a valid positive integer.'''
    res = int_coerce(arg)
    return res if res is nil or res >= 0 else nil


def create_int_range_coerce(min, max):
    '''Create a coercer to check integers between a range.'''
    min, max = vouch(int_coerce, min), vouch(int_coerce, max)
    if min < max:
        @coercer
        def inner(arg):
            'Check if `arg` is a valid integer between "{}" and "{}".'
            arg = int_coerce(arg)
            if t(arg) and min <= arg <= max:
                return arg
            else:
                return nil
        inner.__name__ = str('int_between_{}_and_{}_coerce'.format(min, max))
        inner.__doc__ = inner.__doc__.format(min, max)
        return inner
    else:
        msg = '"{}" must be less than or equal "{}"'
        raise ValueError(msg.format(min, max))


# Identifiers and strings

# TODO: In Py3k "ña" is a valid identifier and this regex won't allow it
_IDENTIFIER_REGEX = re.compile(r'(?i)^[_a-z][\w]*$')


# XXX: 'eight' pending.
@coercer
def identifier_coerce(arg):
    '''Check if `arg` is a valid Python identifier.

    .. note:: Only Python 2's version of valid identifier. This means that
              some Python 3 valid identifiers are not considered valid.  This
              helps to keep things working the same in Python 2 and 3.

    '''
    ok = isinstance(arg, str) and _IDENTIFIER_REGEX.match(arg)
    return str(arg) if ok else nil


_FULL_IDENTIFIER_REGEX = re.compile(r'(?i)^[_a-z][\w]*([.][_a-z][\w]*)*$')


@coercer
def full_identifier_coerce(arg):
    '''Check if `arg` is a valid dotted Python identifier.

    See `identifier_coerce`:func: for what "validity" means.

    '''
    ok = isinstance(arg, str) and _FULL_IDENTIFIER_REGEX.match(arg)
    return str(arg) if ok else nil


@coercer
def names_coerce(arg):
    '''Check `arg` as a tuple of valid object names (identifiers).

    If only one string is given, is returned as the only member of the
    resulting tuple.

    '''
    arg = (arg,) if isinstance(arg, str) else tuple(arg)
    return iterable(identifier_coerce)(arg)


# == Iterators ==

def create_unique_member_coerce(coerce, container):
    '''Useful to wrap member coercers when coercing containers.

    See `iterable`:class: and `mapping`:class:.

    Resulting coercer check that a member must be unique (not repeated) after
    it's coerced.

    For example::

      >>> from xotl.tools.values import (mapping, create_unique_member_coerce,
      ...                            int_coerce, float_coerce)

      >>> sample = {'1': 1, 2.0: '3', 1.0 + 0j: '4.1'}

      >>> dc = mapping(int_coerce, float_coerce)
      >>> dc(dict(sample))
      {1: 1.0, 2: 3.0}

      >>> dc = mapping(create_unique_member_coerce(int_coerce), float_coerce)
      >>> dc(dict(sample))
      nil

    '''
    coerce = vouch(coercer, coerce)

    @coercer
    def inner(arg):
        '''Check a member with "{}" coercer and warrant that is unique.'''
        # assert arg in container
        res = coerce(arg)
        if t(res) and hash(res) != hash(arg) and res in container:
            res = nil
        return res

    cname = coercer_name(coerce)
    inner.__name__ = str('unique_member_{}_coerce'.format(cname))
    inner.__doc__ = inner.__doc__.format(cname)
    return inner


@coercer
def sized_coerce(arg):
    '''Return a valid sized iterable from `arg`.

    If `arg` is iterable but not sized, is converted to a list.  For example::

      >>> sized_coerce(i for i in range(1, 10, 2))
      [1, 3, 5, 7, 9]

      >>> s = {1, 2, 3}
      >>> sized_coerce(s) is s
      True

    '''
    from collections import Iterable, Sized
    if isinstance(arg, Iterable):
        return arg if isinstance(arg, Sized) else list(arg)
    else:
        return nil


@coercer.register
class custom:
    '''Base class for any custom coercer.

    The field `inner` stores an internal data used for the custom coercer;
    could be a callable, an inner coercer, or a tuple of inner checkers if
    more than one is needed, ...

    The field `scope` stores the exit (not regular) condition: the value that
    fails or -if needed- a tuple with (exit-value, exit-coercer) or
    (error-value, error).  The exit condition is not always a failure, for
    example in `some`:class: it is the one that is valid among other inner
    coercers.  To understand better this think on (AND, OR) operators a chain
    of ANDs exits with the first failure and a chains of ORs exits with the
    first success.

    All custom coercers are callable (must redefine `__call__`:meth:)
    receiving one argument that must be coerced.  For example::

      >>> def foobar(*args):
      ...     coerce = pargs(int_coerce)
      ...     return coerce(args)

    This class has two protected fields (`_str_join` and `_repr_join`) that
    are used to call `coercer_name`:func: in `__str__`:meth: and
    `__repr__`:meth: special methods.

    '''
    __slots__ = ('inner', 'scope')

    _str_join = '_'
    _repr_join = ', '

    def __init__(self, *args, **kwargs):
        # This constructor is a placeholder for those custom coercers that can
        # return an instance of a different type in the `__new__`:meth:.
        self.scope = Unset

    def __str__(self):
        name = coercer_name(self.inner, join=self._str_join)
        cls_name = type(self).__name__
        return str('{}_{}_coerce'.format(name, cls_name))

    def __repr__(self):
        name = coercer_name(self.inner, join=self._repr_join)
        cls_name = type(self).__name__
        return str('{}({})'.format(cls_name, name))

    def __call__(self, arg):
        return nil

    @classmethod
    def flatten(cls, obj, avoid=Unset):
        '''Flatten a coercer set.

        :param obj: Could be a coercer representing other inner coercers, or a
               tuple or list containing coercers.

        '''
        aux = obj.inner if isinstance(obj, cls) else obj
        if isinstance(aux, (tuple, list)):
            if not types_tuple_coerce(aux):
                res = (i for l in map(cls.flatten, aux) for i in l)
            else:
                res = (coercer(aux),)
        else:
            res = (aux,)
        if avoid is not Unset:
            res = (i for i in res if i is not avoid)
        return tuple(res)


class istype(custom):
    '''Pure type-checker.

    It's constructed from an argument valid for `types_tuple_coerce`:func:
    coercer.

    For example::

        >>> int_coerce = istype(int)

        >>> int_coerce(1)
        1

        >>> int_coerce('1')
        nil

        >>> number_coerce = istype((int, float, complex))

        >>> number_coerce(1.25)
        1.25

        >>> number_coerce('1.25')
        nil

    '''
    __slots__ = ()

    def __new__(cls, types):
        if types:
            self = super().__new__(cls)
            self.inner = vouch(types_tuple_coerce, types)
            return self
        else:
            return void_coerce

    def __call__(self, arg):
        return arg if isinstance(arg, self.inner) else nil


class typecast(istype):
    '''A type-caster.

    It's constructed from an argument valid for `types_tuple_coerce`:func:
    coercer.  Similar to `istype`:class: but try to convert the value if
    needed.

    For example::

        >>> int_cast = typecast(int)

        >>> int_cast('1')
        1

        >>> int_cast('1x')
        nil

    '''
    __slots__ = ()

    def __call__(self, arg):
        res = super().__call__(arg)
        i = 0
        while not t(res) and i < len(self.inner):
            try:
                tp = self.inner[i]
                res = tp(arg)
                self.scope = tp
            except Exception:
                i += 1
        return res


class safe(custom):
    '''Uses a function (or callable) in a safe way.

    Receives a coercer that expects only one argument and returns another
    value.

    If the returned value is a ``boolean`` (maybe the coercer is a predicate),
    it's converted to a ``logical`` instance.

    The wrapped coercer is called in a safe way (inside try/except); if an
    exception is raised the coercer returns ``nil`` and the error is saved in
    the instance attribute ``scope``.

    '''
    __slots__ = ()

    def __init__(self, func):
        super().__init__()
        self.inner = vouch(coercer, func)

    def __call__(self, arg):
        try:
            from xotl.tools.symbol import boolean
            res = self.inner(arg)
            return logical(res) if isinstance(res, boolean) else res
        except Exception as error:
            self.scope = (arg, error)
            return nil


class compose(custom):
    '''Returns the composition of several inner `coercers`.

    ``compose(f1, ... fn)`` is equivalent to f1(...(fn(arg))...)``.

    If no coercer is given return `identity_coerce`:func:.

    Could be considered an "AND" operator with some light differences because
    the nature of coercers: ordering the coercers is important when some can
    modify (adapt) original values.

    If no value results in `coercers`, a default coercer could be given as a
    keyword argument; `identity_coerce` is assumed if missing.

    '''
    __slots__ = ()

    def __new__(cls, *coercers, **kwds):
        inner = cls.flatten(coercers, avoid=identity_coerce)
        count = len(inner)
        if count > 1:
            self = super().__new__(cls)
            self.inner = inner
            return self
        elif count == 1:
            return inner[0]
        else:
            res = kwds.pop('default', identity_coerce)
            if not kwds:
                return res
            else:
                msg = '`compose` got unexpected keyword argument(s) "{}"'
                raise TypeError(msg.format(set(kwds)))

    def __call__(self, arg):
        coercers = self.inner
        i = 0
        res = arg
        ok = True
        while ok and i < len(coercers):
            coerce = coercers[i]
            aux = coerce(res)
            if t(aux):
                i += 1
            else:
                ok = False
                self.scope = (res, coerce)
            res = aux
        return res


class some(custom):
    '''Represent OR composition of several inner `coercers`.

    ``compose(f1, ... fn)`` is equivalent to f1(arg) or f2(arg) ... fn(arg)``
    in the sense "the first not `nil`".

    If no coercer is given return `void_coerce`:func:.

    '''
    __slots__ = ()

    def __new__(cls, *coercers):
        inner = cls.flatten(coercers, avoid=void_coerce)
        if len(inner) > 1:
            self = super().__new__(cls)
            self.inner = inner
            return self
        elif len(inner) == 1:
            return inner[0]
        else:
            return void_coerce

    def __call__(self, arg):
        coercers = self.inner
        i = 0
        res = nil
        while res is nil and i < len(coercers):
            coercer = coercers[i]
            value = coercer(arg)
            if t(value):
                res = value
                self.scope = coercer
            else:
                i += 1
        return res


class combo(custom):
    '''Represent a zip composition of several inner `coercers`.

    An instance of this class is constructed from a sequence of coercers and
    the its purpose is coerce a sequence of values.  Return a sequence\
    [#type]_ where each item contains the i-th element from applying the i-th
    coercer to the i-th value from argument sequence::

      coercers -> (coercer-1, coercer-2, ... )
      values -> (value-1, value-2, ... )
      combo(coercers)(values) -> (coercer-1(value-1), coercer-2(value-2), ...)

    If any value is coerced invalid, the function returns `nil` and the
    combo's instance variable `scope` receives the duple ``(failed-value,
    failed-coercer)``.

    The returned sequence is truncated in length to the length of the shortest
    sequence (coercers or arguments).

    If no coercer is given, all sequences are coerced as empty.

    .. [#type] The returned sequence is of the same type as the argument
               sequence if possible.

    '''
    __slots__ = ()

    def __init__(self, *coercers):
        super().__init__()
        coercers = pargs(coercer)(coercers)
        self.inner = tuple(vouch(coercer, c) for c in coercers)

    def __call__(self, arg):
        from collections import Iterable
        if isinstance(arg, Iterable):
            coercers = self.inner
            items = iter(arg)
            i = 0
            res = []
            ok = True
            while t(res) and ok and i < len(coercers):
                item = next(items, Unset)
                if item is not Unset:
                    coerce = coercers[i]
                    value = coerce(item)
                    if t(value):
                        res.append(value)
                        i += 1
                    else:
                        res = nil
                        self.scope = (item, coerce)
                else:
                    ok = False
            if t(res):
                try:
                    res = type(arg)(res)
                except Exception:
                    pass
        else:
            res = nil
        return res


class pargs(custom):
    r'''Create a inner coercer that check variable argument passing.

    Created coercer closure must always receives an argument that is an valid
    iterable with all members coerced properly with the argument of this outer
    creator function.

    If the inner closure argument has only a member and this one is not
    properly coerced but it's an iterabled with all members that coerced well,
    this member will be the assumed iterable instead the original argument.

    In the following example::

      >>> from xotl.tools.values import (iterable, int_coerce)

      >>> def foobar(*args):
      ...     coerce = iterable(int_coerce)
      ...     return coerce(args)

      >>> args = (1, 2.0, '3.0')
      >>> foobar(*args)
      (1, 2, 3)

      >>> foobar(args)
      nil

    An example using `pargs`:class:\ ::

      >>> from xotl.tools.values import (pargs, int_coerce)

      >>> def foobar(*args):
      ...     # Below, "coercer" receives the returned "inner"
      ...     coerce = pargs(int_coerce)
      ...     return coerce(args)

      >>> args = (1, 2.0, '3.0')
      >>> foobar(*args)
      (1, 2, 3)

      >>> foobar(args)
      (1, 2, 3)

    The second form is an example of the real utility of this coercer
    closure: if by error a sequence is passed as it to a function that
    expect a variable number of argument, this coercer fixes it.

    Instance variable `scope` stores the last processed invalid argument.

    When executed, usually `arg` is a tuple received by a function as
    ``*args`` form.

    When executed, returns a tuple, or the same type of source iterable
    argument if possible.

    See `xotl.tools.params`:mod: for a more specialized and full function
    arguments conformer.

    See `combo`:class: for a combined coercer that  validate each member with
    a separate member coercer.

    '''
    __slots__ = ()

    def __init__(self, arg_coerce):
        super().__init__()
        self.inner = vouch(coercer, arg_coerce)

    def __call__(self, arg):
        from collections import Iterable
        coerce = self.inner
        if isinstance(arg, Iterable):
            arg = tuple(arg)
            if len(arg) == 1:
                item = arg[0]
                aux = coerce(item)
                if t(aux):
                    res = (aux, )
                elif isinstance(item, Iterable):
                    res = Unset
                    arg = tuple(item)
                else:
                    self.scope = item
                    res = nil
            else:
                res = Unset
            if res is Unset:
                res = arg
                i = 0
                while t(res) and i < len(res):
                    item = res[i]
                    new = coerce(item)
                    if t(new):
                        if new is not item:
                            if isinstance(res, tuple):
                                res = list(res)
                            res[i] = new
                        i += 1
                    else:
                        self.scope = item
                        res = nil
                if t(res):
                    res = tuple(res)
        else:
            res = nil
        return res


class iterable(custom):
    '''Create a inner coercer that coerces an `iterable` member a member.

    See constructor for more information.

    Return a list, or the same type of source iterable argument if possible.

    For example::

      >>> from xotl.tools.values import (iterable, int_coerce,
      ...                            create_unique_member_coerce)

      >>> sample = {'1', 1, '1.0'}

      >>> sc = iterable(int_coerce)
      >>> sc(set(sample)) == {1}
      True

    See `mapping`:class: for more details of this problem.  The equivalent
    safe example is::

      >>> member_coerce = create_unique_member_coerce(int_coerce, sample)
      >>> sc = iterable(member_coerce)
      >>> sc(set(sample))
      nil

    when executed coerces `arg` (an iterable) member a member using
    `member_coercer`. If any member coercion fails, the full execution also
    fails.

    There are three types of results when an instance is executed:
    (1) iterables that are coerced without modifications, (2) the modified
    ones but conserving its type, and (3) those that are returned in a list.

    '''
    __slots__ = ()

    def __init__(self, member_coerce, outer_coerce=True):
        '''Constructor for `iterable`:class: coercers.

        :param member_coerce: A coerce to check each iterable member.

        :param outer_coerce: A coerce to check the type of the entire
               iterable.  Normally a type or tuple of types subclases of
               ``collections.Iterable``.

        '''
        super().__init__()
        member_coerce = vouch(coercer, member_coerce)
        outer_coerce = compose(coercer(outer_coerce), sized_coerce)
        self.inner = (member_coerce, outer_coerce)

    def __call__(self, arg):
        from collections import (Set, Sequence, MutableSequence)
        member_coerce, outer_coerce = self.inner
        modified = False
        aux = outer_coerce(arg)
        if t(aux):
            arg = aux
            if isinstance(arg, Sequence):
                res = arg
                retyped = False
                mutable = isinstance(arg, MutableSequence)
            else:
                res = list(arg)
                retyped = mutable = True
            i = 0
            while t(res) and i < len(res):
                item = res[i]
                new = member_coerce(item)
                if t(new):
                    if new is not item:
                        if not mutable:
                            res = list(res)
                            retyped = mutable = True
                        res[i] = new
                        modified = True
                    i += 1
                else:
                    self.scope = item
                    res = nil
            if t(res):
                if isinstance(arg, Set) and not modified:
                    res = arg
                elif retyped:
                    try:
                        res = type(arg)(res)
                    except Exception:
                        pass
        else:
            self.scope = arg
            res = nil
        return res


class mapping(custom):
    '''Create a coercer to check dictionaries.

    Receives two coercers, one for keys and one for values.

    For example::

      >>> from xotl.tools.values import (mapping, int_coerce, float_coerce,
      ...                                create_unique_member_coerce)

      >>> sample = {'1': 1, 2.0: '3', 1.0 + 0j: '4.1'}

      >>> dc = mapping(int_coerce, float_coerce)
      >>> dc(dict(sample)) == {1: 1.0, 2: 3.0}
      True

    When coercing containers it's probable that members become repeated after
    coercing them.  This could be not desirable (mainly in sets and
    dictionaries). In those cases use `create_unique_member_coerce`:func: to
    wrap member coercer.  For example::

      >>> key_coerce = create_unique_member_coerce(int_coerce, sample)
      >>> dc = mapping(key_coerce, float_coerce)
      >>> dc(dict(sample))
      nil

    Above problem is because it's the same integer (same hash) coerced
    versions of ``'1'`` and ``1.0+0j``.

    This problem of objects of different types that have the same hash is a
    problem to use a example as below::

      >>> {1: int, 1.0: float, 1+0j: complex} == {1: complex}
      True

    '''
    __slots__ = ()
    _str_join = _repr_join = ':'

    def __new__(cls, key_coercer=Unset, value_coercer=Unset):
        '''Constructor for `mapping`:class: coercers.

        :param key_coercer: A coerce to check each one of the mapping keys.

        :param value_coercer: A coerce to check each one of corresponding
               mapping values.

        '''
        from collections import Mapping
        if key_coercer is value_coercer is Unset:
            return coercer(Mapping)
        else:
            self = super().__new__(cls)
            key_coercer = vouch(coercer, key_coercer or True)
            value_coercer = vouch(coercer, value_coercer or True)
            self.inner = (key_coercer, value_coercer)
            return self

    def __call__(self, arg):
        from collections import (Mapping, MutableMapping)
        if isinstance(arg, Mapping):
            key_coercer, value_coercer = self.inner
            res = arg
            retyped = False
            mutable = isinstance(arg, MutableMapping)
            keys = list(res)
            i = 0
            while t(res) and i < len(keys):
                key = keys[i]
                value = res[key]
                new_key = key_coercer(key)
                if t(new_key):
                    new_value = value_coercer(value)
                    if t(new_value):
                        if new_key is not key or new_value is not value:
                            if not mutable:
                                res = dict(res)
                                retyped = mutable = True
                            if key is not new_key:
                                del res[key]
                            res[new_key] = new_value
                        i += 1
                    else:
                        self.scope = ({key: value}, value_coercer)
                        res = nil
                else:
                    self.scope = ({key: value}, key_coercer)
                    res = nil
            if t(res) and retyped:
                try:
                    res = type(arg)(res)
                except Exception:
                    pass
        else:
            self.scope = ()
            res = nil
        return res


del re, ABCMeta, lwraps



from datetime import *
from typing import *

def new_date(d: date) -> date: ...
def new_datetime(d: date) -> datetime: ...

def strfdelta(delta: timedelta) -> str: ...
def strftime(dt: date, fmt: str) -> str: ...

def parse_date(value: str = None) -> date: ...
def parse_datetime(value: str = None) -> datetime: ...

def get_month_first(ref: date = None) -> date: ...
def get_month_last(ref: date = None) -> date: ...
def get_next_month(ref: date = None, lastday: bool = False) -> date: ...

def is_full_month(start: date, end: date) -> bool: ...

class flextime(timedelta):
    @classmethod
    def parse_simple_timeformat(cls, which: str) -> Tuple[int, int, int]: ...


class TimeSpan:
    start_date: Optional[date]
    end_date: Optional[date]

    def __init__(self, start_date: Union[str, date] = None,
                 end_date: Union[str, date] = None) -> None: ...

    @classmethod
    def from_date(self, date: date) -> 'TimeSpan': ...

    @property
    def past_unbound(self) -> bool: ...

    @property
    def future_unbound(self) -> bool: ...


    @property
    def unbound(self) -> bool: ...

    @property
    def bound(self) -> bool: ...

    @property
    def valid(self) -> bool: ...

    def __contains__(self, other: date) -> bool: ...
    def overlaps(self, other: 'TimeSpan') -> bool: ...
    def isdisjoint(self, other: 'TimeSpan') -> bool: ...
    def __le__(self, other: 'TimeSpan') -> bool: ...
    def issubset(self, other: 'TimeSpan') -> bool: ...
    def __lt__(self, other: 'TimeSpan') -> bool: ...
    def __gt__(self, other: 'TimeSpan') -> bool: ...
    def __ge__(self, other: 'TimeSpan') -> bool: ...
    def covers(self, other: 'TimeSpan') -> bool: ...
    def issuperset(self, other: 'TimeSpan') -> bool: ...

    def __iter__(self) -> Iterator[date]: ...
    def __getitem__(self, index:int) -> date: ...

    def __eq__(self, other: Union[date, 'TimeSpan']) -> bool: ...
    def __and__(self, other: 'TimeSpan') -> 'TimeSpan': ...
    def __lshift__(self, delta: Union[int, timedelta]) -> 'TimeSpan': ...
    def __rshift__(self, delta: Union[int, timedelta]) -> 'TimeSpan': ...

    def intersection(self, *others: 'TimeSpan') -> 'TimeSpan': ...

    def diff(self, other: 'TimeSpan') -> Tuple[TimeSpan, TimeSpan]: ...


class DateTimeSpan(TimeSpan):
    start_datetime: Optional[datetime]
    end_datetime: Optional[datetime]

    def __init__(self, start_datetime: Union[str, date] = None,
                 end_datetime: Union[str, date] = None) -> None: ...

    @classmethod
    def from_date(self, d: date) -> 'DateTimeSpan': ...

    @classmethod
    def from_datetime(self, dt: datetime) -> 'DateTimeSpan': ...

    @classmethod
    def from_timespan(self, ts: TimeSpan) -> 'DateTimeSpan': ...

    def __iter__(self) -> Iterator[datetime]: ...
    def __getitem__(self, index: int) -> datetime: ...

    def __and__(self, other: TimeSpan) -> 'DateTimeSpan': ...
    def __lshift__(self, delta: Union[int, timedelta]) -> 'DateTimeSpan': ...
    def __rshift__(self, delta: Union[int, timedelta]) -> 'DateTimeSpan': ...
    def intersection(self, *others: TimeSpan) -> 'DateTimeSpan': ...

    def diff(self, other: TimeSpan) -> Tuple['DateTimeSpan', 'DateTimeSpan']: ...


EmptyTimeSpan: DateTimeSpan



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
'''Functional tools for functions that returns iterators (generators, etc.)

.. warning:: This module is experimental.  It may be removed completely, moved
   or otherwise changed.

'''
from functools import reduce
from xotl.tools.deprecation import deprecated_alias


def kleisli_compose(*fs):
    '''The Kleisli composition operator.

    For two functions, ``kleisli_compose(g, f)`` returns::

       lambda x: (z for y in f(x) for z in g(y))

    In general this is, ``reduce(_compose, fs, lambda x: [x])``; where
    ``_compose`` is the lambda for two arguments.

    .. note:: Despite name (Kleisli), Python does not have a true Monad_
       type-class.  So this function works with functions taking a single
       argument and returning an iterator -- it also works with iterables.

    .. _Monad: https://en.wikipedia.org/wiki/Monad_(functional_programming)

    .. versionadded:: 1.9.6
    .. versionchanged:: 1.9.7 Name changed to ``kleisli_compose``.

    '''
    def _kleisli_compose(g, f):
        # (>>.) :: Monad m => (b -> m c) -> (a -> m b) -> a -> m c
        # g >>. f = \x -> f x >>= g
        #
        # In the list monad:
        #
        # g >>. f = \x -> concat (map g (f x))
        return lambda x: (z for y in f(x) for z in g(y))

    if len(fs) == 2:
        # optimize a bit so that we can avoid the 'lambda x: [x]' for common
        # cases.
        return _kleisli_compose(*fs)
    else:
        return reduce(_kleisli_compose, fs, lambda x: iter([x]))


iter_compose = deprecated_alias(kleisli_compose)


del deprecated_alias



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# flake8:  noqa
import os
import sys
import versioneer

from setuptools import setup, find_packages

try:
    execfile
except NameError:

    def execfile(filename):
        "To run in Python 3"
        import builtins

        exec_ = getattr(builtins, "exec")
        with open(filename, "rb") as f:
            code = compile(f.read().decode("utf-8"), filename, "exec")
            return exec_(code, globals())


# Import the version from the release module
project_name = os.environ.get("PROJECT_NAME", "xotl.tools")
_current_dir = os.path.dirname(os.path.abspath(__file__))
dev_classifier = "Development Status :: 5 - Production/Stable"

from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.test_args)
        sys.exit(errno)


_cmdclass = {"test": PyTest}
_cmdclass.update(versioneer.get_cmdclass())


setup(
    name=project_name,
    version=versioneer.get_version(),
    description=("Collection of usefull algorithms and other very " "disparate stuff"),
    long_description=open(os.path.join(_current_dir, "README.rst")).read(),
    classifiers=[
        # Get from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        dev_classifier,
        "Intended Audience :: Developers",
        (
            "License :: OSI Approved :: "
            "GNU General Public License v3 or later (GPLv3+)"
        ),
        "Operating System :: POSIX :: Linux",  # This is where we are
        # testing. Don't promise
        # anything else.
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="",
    author="Merchise",
    author_email="project+xoutil@merchise.org",
    # TODO: @med, @manu Negotiate with Maykel Moya to obtain plain "Merchise"
    #       folder at "https://github.com"
    url="https://github.com/merchise-autrement/xoutil/",
    license="GPLv3+",
    tests_require=["pytest"],
    packages=find_packages(exclude=["ez_setup", "examples", "tests"]),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.4",
    install_requires=[
        'monotonic; python_version<"3.3"',
        'contextlib2; python_version<"3.4"',
        'typing; python_version<"3.5"',
    ],
    extras_require={"recommended": ["python-dateutil", 'enum34; python_version<"3.4"']},
    cmdclass=_cmdclass,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Task, multitasking, and concurrent programming tools.

.. warning:: Experimental.  API is not settled.

.. versionadded:: 1.8.0

'''

import sys
from xotl.tools.deprecation import deprecated_alias


# TODO: Must be implemented using `xotl.tools.api` mechanisms for correct
# driver determination, in this case "thread-local data".
if 'greenlet' in sys.modules:
    from ._greenlet_local import local    # noqa
else:
    try:
        from threading import local    # noqa
    except ImportError:
        from dummy_threading import local    # noqa

del sys


class AutoLocal(local):
    '''Initialize thread-safe local data in one shoot.

    Typical use is::

        >>> from xotl.tools.tasking import AutoLocal
        >>> context = AutoLocal(cause=None, traceback=None)

    When at least one attribute is given, ``del AutoLocal`` it's executed
    automatically avoiding this common statement at the end of your module.

    '''
    def __init__(self, **attrs):
        import sys
        super().__init__()
        for attr in attrs:
            setattr(self, attr, attrs[attr])
        if attrs:
            g = sys._getframe(1).f_globals
            tname = type(self).__name__
            if tname in g:
                del g[tname]


#: The minimal time (in seconds) to wait between retries.
#:
#: .. versionadded:: 1.8.2
MIN_WAIT_INTERVAL = 20 / 1000  # 20 ms

#: The default time (in seconds) to wait between retries.
#:
#: .. versionadded:: 1.8.2
DEFAULT_WAIT_INTERVAL = 50 / 1000  # 50 ms


class ConstantWait:
    '''A constant wait algorithm.

    Instances are callables that comply with the need of the `wait` argument
    for `retrier`:class:.  This callable always return the same `wait` value.

    We never wait less than `MIN_WAIT_INTERVAL`:data:.

    .. versionadded:: 1.8.2

    .. versionchanged:: 1.9.1 Renamed; it was ``StandardWait``.  The old name
       is kept as a deprecated alias.

    .. versionchanged:: 2.0.1 Renamed; it was ``StandardWait``.  The old name
       is kept as a deprecated alias.

    '''
    def __init__(self, wait=DEFAULT_WAIT_INTERVAL):
        import numbers
        if not isinstance(wait, numbers.Real):
            raise TypeError("'wait' must a number.")
        self.wait = max(MIN_WAIT_INTERVAL, wait)

    def __call__(self, prev=None):
        return self.wait


StandardWait = deprecated_alias(
    ConstantWait,
    msg='StandardWait is deprecated. Use ConstantWait instead'
)


class BackoffWait:
    '''A wait algorithm with an exponential backoff.

    Instances are callables that comply with the need of the `wait` argument
    for `retrier`:class:.

    At each call the wait is increased by doubling `backoff` (given in
    milliseconds).

    We never wait less than `MIN_WAIT_INTERVAL`:data:.

    .. versionadded:: 1.8.2

    '''
    def __init__(self, wait=DEFAULT_WAIT_INTERVAL, backoff=1):
        self.wait = max(MIN_WAIT_INTERVAL, wait)
        self.backoff = min(max(0.1, backoff), 1)

    def __call__(self, prev=None):
        res = self.wait + (self.backoff / 1000)
        self.backoff = self.backoff * 2
        return res


def get_backoff_wait(n, *, wait=DEFAULT_WAIT_INTERVAL, backoff=1):
    '''Compute the total backoff wait time after `n` tries.

    .. seealso:: `BackoffWait`:class:.

    .. versionadded:: 2.1.0

    '''
    res = 0
    fn = BackoffWait(wait=wait, backoff=backoff)
    for _ in range(n):
        res = fn(prev=res)
    return res


def retry(fn, args=None, kwargs=None, *, max_tries=None, max_time=None,
          wait=DEFAULT_WAIT_INTERVAL, retry_only=None):
    '''Run `fn` with args and kwargs in an auto-retrying loop.

    See `retrier`:class:.  This is just::

       >>> retrier(max_tries=max_tries, max_time=max_time, wait=wait,
       ...         retry_only=retry_only)(fn, *args, **kwargs)

    .. versionadded:: 1.8.2

    '''
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}
    return retrier(max_tries=max_tries, max_time=max_time, wait=wait,
                   retry_only=retry_only)(fn, *args, **kwargs)


class retrier:
    '''An auto-retrying dispatcher.

    A retrier it's a callable that takes another callable (`func`) and its
    arguments and runs it in an auto-retrying loop.

    If `func` raises any exception that is in `retry_only`, and it has being
    tried less than `max_tries` and the time spent executing the function
    (waiting included) has not reached `max_time`, the function will be
    retried.

    `wait` can be a callable or a number.  If `wait` is callable, it must take
    a single argument with the previous waiting we did (`None`:data: for the
    first retry) and return the number of seconds to wait before retrying.

    If `wait` is a number, we convert it to a callable with
    `ConstantWait(wait) <ConstantWait>`:class:.

    .. seealso:: `BackoffWait`:class:

    If `retry_only` is None, all exceptions (that inherits from Exception)
    will be retried.  Otherwise, only the exceptions in `retry_only` will be
    retried.

    Waiting is done with `time.sleep`:func:.  Time tracking is done with
    `time.monotonic`:func:.

    .. versionadded:: 1.8.2

    '''
    def __init__(self, max_tries=None, max_time=None,
                 wait=DEFAULT_WAIT_INTERVAL, retry_only=None):
        if not max_tries and not max_time:
            raise TypeError('One of tries or times must be set')
        self.max_tries = max_tries
        self.max_time = max_time
        if not callable(wait):
            self.wait = ConstantWait(wait)
        else:
            self.wait = wait
        if not retry_only:
            self.retry_only = (Exception, )
        else:
            self.retry_only = retry_only

    def __call__(self, fn, *args, **kwargs):
        return self.decorate(fn)(*args, **kwargs)

    def decorate(self, fn):
        '''Return `fn` decorated to run in an auto-retry loop.

        You can use this to decorate a function you'll always run inside a
        retrying loop:

           >>> @retrier(max_tries=5, retry_only=TransientError).decorate
           ... def read_from_url(url):
           ...     pass

        '''
        from time import monotonic as clock, sleep
        from xotl.tools.future.functools import wraps
        max_time = self.max_time
        max_tries = self.max_tries

        @wraps(fn)
        def inner(*args, **kwargs):
            t = 0
            done = False
            start = clock()
            waited = None
            while not done:
                try:
                    return fn(*args, **kwargs)
                except self.retry_only:
                    t += 1
                    reached_max_tries = max_tries and t >= max_tries
                    max_time_elapsed = max_time and clock() - start >= max_time
                    retry = not reached_max_tries and not max_time_elapsed
                    if retry:
                        waited = self.wait(waited)
                        sleep(waited)
                    else:
                        raise

        return inner


del deprecated_alias



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Extends the standard `datetime` module.

- Python's ``datetime.strftime`` doesn't handle dates previous to 1900.
  This module define classes to override `date` and `datetime` to support the
  formatting of a date through its full proleptic Gregorian date range.

Based on code submitted to comp.lang.python by Andrew Dalke, copied from
Django and generalized.

You may use this module as a drop-in replacement of the standard library
`datetime` module.

'''

from datetime import *    # noqa
from datetime import timedelta
import datetime as _stdlib    # noqa

from re import compile as _regex_compile
from time import strftime as _time_strftime

from enum import IntEnum
from typing import Iterator, Tuple, Union  # noqa

from xotl.tools.deprecation import deprecated


class WEEKDAY(IntEnum):
    '''Simple constants for 'weekday' method.'''
    MONDAY = 0
    TUESDAY = 1
    WEDNESDAY = 2
    THURSDAY = 3
    FRIDAY = 4
    SATURDAY = 5
    SUNDAY = 6


class ISOWEEKDAY(IntEnum):
    '''Simple constants for 'weekday' method.'''
    MONDAY = 1
    TUESDAY = 2
    WEDNESDAY = 3
    THURSDAY = 4
    FRIDAY = 5
    SATURDAY = 6
    SUNDAY = 7


@deprecated("plain objects")
def assure(obj):
    '''Make sure that a `date` or `datetime` instance is a safe version.

    This is only a type checker alternative to standard library.

    '''
    if isinstance(obj, (date, datetime, _stdlib.time, timedelta)):
        return obj
    else:
        raise TypeError('Not valid type for datetime assuring: %s' % obj)


@deprecated(assure)
def new_date(d):
    '''Generate a safe date from a legacy datetime date object.'''
    return date(d.year, d.month, d.day)


@deprecated(assure)
def new_datetime(d):
    '''Generate a safe datetime given a legacy date or datetime object.'''
    args = [d.year, d.month, d.day]
    if isinstance(d, datetime):
        args.extend([d.hour, d.minute, d.second, d.microsecond, d.tzinfo])
    return datetime(*args)


del deprecated


# This library does not support strftime's "%s" or "%y" format strings.
# Allowed if there's an even number of "%"s because they are escaped.
_illegal_formatting = _regex_compile(br"((^|[^%])(%%)*%[sy])")


def _year_find_all(fmt, year, no_year_tuple):
    text = _time_strftime(fmt, (year,) + no_year_tuple)
    regex = _regex_compile(str(year))
    return {match.start() for match in regex.finditer(text)}


_TD_LABELS = 'dhms'    # days, hours, minutes, seconds


def _strfnumber(number, format_spec='%0.2f'):
    '''Convert a floating point number into string using a smart way.

    Used internally in strfdelta.

    '''
    res = format_spec % number
    if '.' in res:
        res = res.rstrip('0')
        if res.endswith('.'):
            res = res[:-1]
    return res


def strfdelta(delta):
    '''
    Format a timedelta using a smart pretty algorithm.

    Only two levels of values will be printed.

    ::

        >>> def t(h, m):
        ...     return timedelta(hours=h, minutes=m)

        >>> strfdelta(t(4, 56)) == '4h 56m'
        True

    '''
    ss, sss = str('%s%s'), str(' %s%s')
    if delta.days:
        days = delta.days
        delta -= timedelta(days=days)
        hours = delta.total_seconds() / 60 / 60
        res = ss % (days, _TD_LABELS[0])
        if hours >= 0.01:
            res += sss % (_strfnumber(hours), _TD_LABELS[1])
    else:
        seconds = delta.total_seconds()
        if seconds > 60:
            minutes = seconds / 60
            if minutes > 60:
                hours = int(minutes / 60)
                minutes -= hours * 60
                res = ss % (hours, _TD_LABELS[1])
                if minutes >= 0.01:
                    res += sss % (_strfnumber(minutes), _TD_LABELS[2])
            else:
                minutes = int(minutes)
                seconds -= 60 * minutes
                res = ss % (minutes, _TD_LABELS[2])
                if seconds >= 0.01:
                    res += sss % (_strfnumber(seconds), _TD_LABELS[3])
        else:
            res = ss % (_strfnumber(seconds, '%0.3f'), _TD_LABELS[3])
    return res


def strftime(dt, fmt):
    '''Used as `strftime` method of `date` and `datetime` redefined classes.

    Also could be used with standard instances.

    '''
    if dt.year >= 1900:
        bases = type(dt).mro()
        i = 0
        base = _strftime = type(dt).strftime
        while _strftime == base:
            aux = getattr(bases[i], 'strftime', base)
            if aux != base:
                _strftime = aux
            else:
                i += 1
        return _strftime(dt, fmt)
    else:
        illegal_formatting = _illegal_formatting.search(fmt)
        if illegal_formatting is None:
            year = dt.year
            # For every non-leap year century, advance by 6 years to get into
            # the 28-year repeat cycle
            delta = 2000 - year
            year += 6 * (delta // 100 + delta // 400)
            year += ((2000 - year) // 28) * 28   # Move to around the year 2000
            no_year_tuple = dt.timetuple()[1:]
            sites = _year_find_all(fmt, year, no_year_tuple)
            sites &= _year_find_all(fmt, year + 28, no_year_tuple)
            res = _time_strftime(fmt, (year,) + no_year_tuple)
            syear = "%04d" % dt.year
            for site in sites:
                res = res[:site] + syear + res[site + 4:]
            return res
        else:
            msg = 'strftime of dates before 1900 does not handle  %s'
            raise TypeError(msg % illegal_formatting.group(0))


def parse_date(value=None):
    if value:
        y, m, d = value.split('-')
        return date(int(y), int(m), int(d))
    else:
        return date.today()


def parse_datetime(value=None):
    '''Parse a datime in format 'YYYY-MM-DD HH:MM[:SS][.MS]'.

    The hour-minute component is mandatory.

    '''
    if value:
        d, t = value.split()
        y, m, d = d.split('-')
        if '.' in t:
            moment, ms = t.split('.')
        else:
            moment, ms = t, '0'
        timing = moment.split(':')
        if len(timing) == 2:
            h, mn = timing
            s = 0
        elif len(timing) == 3:
            h, mn, s = timing
        else:
            raise ValueError('Invalid time string %r' % t)
        return datetime(
            int(y), int(m), int(d),
            int(h), int(mn), int(s),
            int(ms)
        )
    else:
        return datetime.now()


def get_month_first(ref=None):
    '''Given a reference date, returns the first date of the same month. If
    `ref` is not given, then uses current date as the reference.
    '''
    aux = ref or date.today()
    y, m = aux.year, aux.month
    return date(y, m, 1)


def get_month_last(ref=None):
    '''Given a reference date, returns the last date of the same month. If
    `ref` is not given, then uses current date as the reference.
    '''
    aux = ref or date.today()
    y, m = aux.year, aux.month
    if m == 12:
        m = 1
        y += 1
    else:
        m += 1
    return date(y, m, 1) - timedelta(1)


def get_next_month(ref=None, lastday=False):
    '''Get the first or last day of the *next month*.

    If `lastday` is False return the first date of the `next month`.
    Otherwise, return the last date.

    The *next month* is computed with regards to a reference date.  If `ref`
    is None, take the current date as the reference.

    Examples:

      >>> get_next_month(date(2017, 1, 23))
      date(2017, 2, 1)

      >>> get_next_month(date(2017, 1, 23), lastday=True)
      date(2017, 2, 28)

    .. versionadded:: 1.7.3

    '''
    result = get_month_last(ref) + timedelta(days=1)
    if lastday:
        return get_month_last(result)
    else:
        return result


def is_full_month(start, end):
    '''Returns true if the arguments comprises a whole month.
    '''
    sd, sm, sy = start.day, start.month, start.year
    em, ey = end.month, end.year
    return ((sd == 1) and (sm == em) and (sy == ey) and
            (em != (end + timedelta(1)).month))


class flextime(timedelta):
    @classmethod
    def parse_simple_timeformat(cls, which):
        if 'h' in which:
            hour, rest = which.split('h')
        else:
            hour, rest = 0, which
        return int(hour), int(rest), 0

    def __new__(cls, *args, **kwargs):
        first = None
        if args:
            first, rest = args[0], args[1:]
        _super = super().__new__
        if first and not rest and not kwargs:
            hour, minutes, seconds = cls.parse_simple_timeformat(first)
            return _super(cls, hours=hour, minutes=minutes, seconds=seconds)
        else:
            return _super(cls, *args, **kwargs)


# TODO: Merge this with the new time span.
def daterange(*args):
    '''Similar to standard 'range' function, but for date objets.

    Returns an iterator that yields each date in the range of ``[start,
    stop)``, not including the stop.

    If `start` is given, it must be a date (or `datetime`) value; and in this
    case only `stop` may be an integer meaning the numbers of days to look
    ahead (or back if `stop` is negative).

    If only `stop` is given, `start` will be the first day of stop's month.

    `step`, if given, should be a non-zero integer meaning the numbers of days
    to jump from one date to the next. It defaults to ``1``. If it's positive
    then `stop` should happen after `start`, otherwise no dates will be
    yielded. If it's negative `stop` should be before `start`.

    As with `range`, `stop` is never included in the yielded dates.

    '''
    import operator
    # Use base classes to allow broader argument values
    from datetime import date, datetime
    if len(args) == 1:
        start, stop, step = None, args[0], None
    elif len(args) == 2:
        start, stop = args
        step = None
    else:
        start, stop, step = args
    if not step and step is not None:
        raise ValueError('Invalid step value %r' % step)
    if not start:
        if not isinstance(stop, (date, datetime)):
            raise TypeError('stop must a date if start is None')
        else:
            start = get_month_first(stop)
    else:
        if stop is not None and not isinstance(stop, (date, datetime)):
            stop = start + timedelta(days=stop)
    if step is None or step > 0:
        compare = operator.lt
    else:
        compare = operator.gt
    step = timedelta(days=(step if step else 1))

    # Encloses the generator so that signature validation exceptions happen
    # without needing to call next().
    def _generator():
        current = start
        while stop is None or compare(current, stop):
            yield current
            current += step
    return _generator()


class DateField:
    '''A simple descriptor for dates.

    Ensures that assigned values must be parseable dates and parses them.

    '''
    def __init__(self, name, nullable=False):
        self.name = name
        self.nullable = nullable

    def __get__(self, instance, owner):
        if instance is not None:
            res = instance.__dict__[self.name]
            return res
        else:
            return self

    def __set__(self, instance, value):
        if value in (None, False):
            # We regard False as None, so that working with Odoo is easier:
            # missing values in Odoo, often come as False instead of None.
            if not self.nullable:
                raise ValueError('Setting None to a required field')
            else:
                value = None
        elif isinstance(value, datetime):
            value = value.date()
        elif not isinstance(value, date):
            value = parse_date(value)
        instance.__dict__[self.name] = value


class DateTimeField(object):
    '''A simple descriptor for datetimes.

    Ensures that assigned values must be parseable date or datetime and parses
    them.

    If `prefer_last_minute` is False when converting from date, the time
    component will be '00:00:00', if True, the time component will be
    '23:59:59'.

    .. versionadded:: 1.9.7

    '''
    def __init__(self, name, nullable=False, prefer_last_minute=False):
        self.name = name
        self.nullable = nullable
        self.prefer_last_minute = prefer_last_minute

    def __get__(self, instance, owner):
        if instance is not None:
            res = instance.__dict__[self.name]
            return res
        else:
            return self

    def __set__(self, instance, value):
        if value in (None, False):
            # We regard False as None, so that working with Odoo is easier:
            # missing values in Odoo, often come as False instead of None.
            if not self.nullable:
                raise ValueError('Setting None to a required field')
            else:
                value = None
        elif isinstance(value, datetime):
            # needed because datetime is subclass of date, and the next
            # condition would match.
            pass
        elif isinstance(value, date):
            if not self.prefer_last_minute:
                value = datetime(value.year, value.month, value.day)
            else:
                value = datetime(value.year, value.month, value.day,
                                 23, 59, 59)
        else:
            try:
                value = parse_datetime(value)
            except ValueError:
                value = parse_date(value)
                self.__set__(instance, value)  # lazy me
                return
        instance.__dict__[self.name] = value


class TimeSpan:
    '''A *continuous* span of time.

    Time spans objects are iterable.  They yield exactly two times: first the
    start date, and then the end date::

       >>> ts = TimeSpan('2017-08-01', '2017-09-01')
       >>> tuple(ts)
       (date(2017, 8, 1), date(2017, 9, 1))

    Time spans objects have two items::

       >>> ts[0]
       date(2017, 8, 1)

       >>> ts[1]
       date(2017, 9, 1)

       >>> ts[:]
       (date(2017, 8, 1), date(2017, 9, 1))

    Two time spans are equal if their start_date and end_date are equal.  When
    comparing a time span with a date, the date is coerced to a time span
    (`from_date`:meth:).

    .. note:: Comparing time spans with date time spans `coerces the time span
              <DateTimeSpan.from_timespan>`:meth: before comparing.

    A time span with its `start` set to None is unbound to the past.  A time
    span with its `end` set to None is unbound to the future.  A time span
    that is both unbound to the past and the future contains all possible
    dates.  A time span that is not unbound in any direction is
    `bound <bound>`:attr:.

    A bound time span is `valid`:attr: if its start date comes before its end
    date.  Unbound time spans are always valid.

    Time spans can `intersect <__mul__>`:meth:, compared for containment of
    dates and by the subset/superset order operations (``<=``, ``>=``).  In
    this regard, they represent the *set* of dates between `start` and `end`,
    inclusively.

    .. warning:: Time spans don't implement the union or difference operations
       expected in sets because the difference/union of two span is not
       necessarily *continuous*.

    '''
    start_date = DateField('start_date', nullable=True)
    end_date = DateField('end_date', nullable=True)

    def __init__(self, start_date=None, end_date=None):
        self.start_date = start_date
        self.end_date = end_date

    @classmethod
    def from_date(self, date: date) -> 'TimeSpan':
        '''Return a new time span that covers a single `date`.'''
        return self(start_date=date, end_date=date)

    @property
    def past_unbound(self) -> bool:
        'True if the time span is not bound into the past.'
        return self.start_date is None

    @property
    def future_unbound(self) -> bool:
        'True if the time span is not bound into the future.'
        return self.end_date is None

    @property
    def unbound(self) -> bool:
        '''True if the time span is `unbound into the past <past_unbound>`:attr: or
        `unbount into the future <future_unbound>`:attr: or both.

        '''
        return self.future_unbound or self.past_unbound

    @property
    def bound(self) -> bool:
        'True if the time span is not `unbound <unbound>`:attr:.'
        return not self.unbound

    @property
    def valid(self) -> bool:
        '''A bound time span is valid if it starts before it ends.

        Unbound time spans are always valid.

        '''
        if self.bound:
            return self.start_date <= self.end_date
        else:
            return True

    def __contains__(self, other):
        # type: (date) -> bool
        '''Test date `other` is in the time span.'''
        if isinstance(other, date):
            if self.start_date and self.end_date:
                return self.start_date <= other <= self.end_date
            elif self.start_date:
                return self.start_date <= other
            elif self.end_date:
                return other <= self.end_date
            else:
                return True
        else:
            return False

    def overlaps(self, other):
        # type: (TimeSpan) -> bool
        '''Test if the time spans overlaps.'''
        return bool(self & other)

    def isdisjoint(self, other):
        # type: (TimeSpan) -> bool
        return not self.overlaps(other)

    def __le__(self, other):
        # type: (TimeSpan) -> bool
        'True if `other` is a superset.'
        return (self & other) == self

    issubset = __le__

    def __lt__(self, other):
        # type: (TimeSpan) -> bool
        'True if `other` is a proper superset.'
        return self != other and self <= other

    def __gt__(self, other):
        # type: (TimeSpan) -> bool
        'True if `other` is a proper subset.'
        return self != other and self >= other

    def __ge__(self, other):
        # type: (TimeSpan) -> bool
        'True if `other` is a subset.'
        # Notice that ge is not the opposite of lt.
        return (self & other) == other

    issuperset = covers = __ge__

    def __iter__(self):
        # type: () -> Iterator[date]
        yield self.start_date
        yield self.end_date

    def __getitem__(self, index):
        # type: (int) -> date
        this = tuple(self)
        return this[index]

    def __eq__(self, other: Union[date, 'TimeSpan']) -> bool:  # type: ignore
        if isinstance(other, date):
            other = type(self).from_date(other)
        elif isinstance(other, DateTimeSpan):
            return other == self
        if not isinstance(other, TimeSpan):
            return NotImplemented
        return (self.start_date == other.start_date and
                self.end_date == other.end_date)

    def __hash__(self):
        return hash((TimeSpan, self.start_date, self.end_date))

    def __and__(self, other):
        # type: (TimeSpan) -> TimeSpan
        '''Get the time span that is the intersection with another time span.

        If two time spans don't overlap, return `EmptyTimeSpan`:data:.

        If `other` is not a TimeSpan we try to create one.  If `other` is a
        date, we create the TimeSpan that starts and end that very day. Other
        types are passed unchanged to the constructor.

        When `other` is a `DateTimeSpan`:class:, convert `self` to a `date
        time span <DateTimeSpan.from_timespan>`:meth: before doing the
        intersection.

        '''
        from xotl.tools.infinity import Infinity
        if isinstance(other, _EmptyTimeSpan):
            return other
        elif isinstance(other, date):
            other = TimeSpan.from_date(other)
        elif isinstance(other, DateTimeSpan):
            return other & self
        elif not isinstance(other, TimeSpan):
            raise TypeError("Invalid type '%s'" % type(other).__name__)
        start = max(
            self.start_date or -Infinity,
            other.start_date or -Infinity
        )
        end = min(
            self.end_date or Infinity,
            other.end_date or Infinity
        )
        if start <= end:
            if start is -Infinity:
                start = None
            if end is Infinity:
                end = None
            return type(self)(start, end)
        else:
            return EmptyTimeSpan
    __mul__ = __rmul__ = __rand__ = __and__

    def __bool__(self):
        return True

    __nonzero__ = __bool__

    def __len__(self):
        '''The amount of dates in the span.

        .. warning:: If the time span is `unbound`:attr: this method returns
                     NotImplemented.  This will make python complain with a
                     TypeError.

        .. versionadded:: 1.8.2

        '''
        if self.bound:
            return (self.end_date - self.start_date).days
        else:
            return NotImplemented

    def __lshift__(self, delta):
        '''Return the time span displaced to the past in `delta`.

        :param delta: The number of days to displace.  It can be either an
                      integer or a `datetime.timedelta`:class:.  The integer
                      will be converted to ``timedelta(days=delta)``.

        .. note:: Delta values that don't amount to at least a day will be the
                  same as 0.

        .. versionadded:: 1.8.2

        .. warning:: Python does have a boundaries for the dates it can
           represent, so displacing a TimeSpan can cause OverflowError.

        '''
        import numbers
        if isinstance(delta, numbers.Integral):
            delta = timedelta(days=delta)   # noqa
        start = self.start_date - delta if self.start_date else None
        end = self.end_date - delta if self.end_date else None
        return type(self)(start, end)

    def __rshift__(self, delta):
        '''Return the time span displaced to the future in `delta`.

        :param delta: The number of days to displace.  It can be either an
                      integer or a `datetime.timedelta`:class:.  The integer
                      will be converted to ``timedelta(days=delta)``.

        .. note:: Delta values that don't amount to at least a day will be the
                  same as 0.

        .. versionadded:: 1.8.2

        .. warning:: Python does have a boundaries for the dates it can
           represent, so displacing a TimeSpan can cause OverflowError.

        '''
        return self << -delta

    def intersection(self, *others):
        'Return ``self [& other1 & ...]``.'
        import operator
        from functools import reduce
        return reduce(operator.mul, others, self)

    def diff(self, other):
        # type: (TimeSpan) -> Tuple[TimeSpan, TimeSpan]
        '''Return the two time spans which (combined) contain all the dates in
        `self` which are not in `other`.

        Notice this method returns a tuple of exactly two items.

        If `other` and `self` don't overlap, return ``(self, EmptyTimeSpan)``.

        If ``self <= other`` is True, return the tuple with the empty time
        span in both positions.

        Otherwise `self` will have some dates which are not in `other`; there
        are possible three cases:

        a) other starts before or at self's start date; return the empty time
           span and the time span containing the dates after `other.end_date`
           up to `self.end_date`

        b) other ends at or after self's end date; return the dates from
           `self.start_date` up to the date before `other.start_date` and the
           empty time span.

        c) `other` is fully contained in `self`; return two non-empty time
           spans as in the previous cases.

        .. versionadded:: 1.9.7

        '''
        if not self & other:
            return self, EmptyTimeSpan  # type: ignore
        other = self & other
        if self == other:
            return EmptyTimeSpan, EmptyTimeSpan  # type: ignore
        else:
            assert self > other
            day = timedelta(days=1)
            if self.start_date == other.start_date:
                return (EmptyTimeSpan,  # type: ignore
                        TimeSpan(other.end_date + day, self.end_date))
            elif self.end_date == other.end_date:
                return (
                    TimeSpan(self.start_date, other.start_date - day),
                    EmptyTimeSpan
                )
            else:
                return (TimeSpan(self.start_date, other.start_date - day),
                        TimeSpan(other.end_date + day, self.end_date))

    def __repr__(self):
        start, end = self
        return 'TimeSpan(%r, %r)' % (start.isoformat() if start else None,
                                     end.isoformat() if end else None)

    __str__ = __repr__


class _EmptyTimeSpan:
    __slots__ = []  # no inner structure

    def __bool__(self):
        return False

    __nonzero__ = __bool__

    def __contains__(self, which):
        return False  # I don't contain noone

    # The empty is equal only to itself
    def __eq__(self, which):
        if isinstance(which, (TimeSpan, date, _EmptyTimeSpan)):
            # We expect `self` to be a singleton, but pickle protocol 1 does
            # not warrant to call our __new__.
            return self is which
        else:
            return NotImplemented

    def __ne__(self, other):
        res = self == other
        if res is not NotImplemented:
            return not res
        else:
            return res

    # The empty set is a subset of any other set.  dates are regarded as the
    # set that contains that
    def __le__(self, which):
        if isinstance(which, (TimeSpan, date, _EmptyTimeSpan)):
            return True
        else:
            return NotImplemented

    # The empty set is only a superset of itself.
    __ge__ = covers = __eq__

    # The empty set is a *proper* subset of any set but itself.  The empty
    # set is disjoint with any other set but itself.
    __lt__ = isdisjoint = __ne__

    # The empty set is a *proper* superset of no one
    def __gt__(self, which):
        if isinstance(which, (TimeSpan, date, _EmptyTimeSpan)):
            return True
        else:
            return NotImplemented

    # `empty | x == empty + x == x`
    def __add__(self, which):
        if isinstance(which, (TimeSpan, date, _EmptyTimeSpan)):
            return which
        else:
            raise TypeError

    __or__ = __add__

    # `empty & x == empty * x == empty`
    def __mul__(self, other):
        if isinstance(other, (TimeSpan, date, _EmptyTimeSpan)):
            return self
        else:
            raise TypeError

    __and__ = __mul__

    def __repr__(self):
        return 'EmptyTimeSpan'

    __str__ = __repr__

    def __new__(cls):
        res = getattr(cls, '_instance', None)
        if res is None:
            res = cls._instance = super().__new__(cls)
        return res

    def __reduce__(self):
        # So that unpickling returns the singleton
        return type(self), ()

    def __len__(self):
        return 0

    def __lshift__(self, delta):
        return self

    def __rshift__(self, delta):
        return self


# I solemnly swear that EmptyTimeSpan is of type DateTimeSpan.
EmptyTimeSpan = _EmptyTimeSpan()


# TODO: Move this to xotl.tools.objects or somewhere else
class SynchronizedField(object):
    '''A synchronized descriptor.

    Whenever the `source` gets updated, update the second.

    '''
    def __init__(self, descriptor, setting_descriptor, set_throu_get=True):
        self.descriptor = descriptor
        self.setting_descriptor = setting_descriptor
        self.set_throu_get = set_throu_get

    def __get__(self, instance, owner):
        return self.descriptor.__get__(instance, owner)

    def __set__(self, instance, value):
        from xotl.tools.context import context
        self.descriptor.__set__(instance, value)
        if (SynchronizedField, self.setting_descriptor) not in context:
            with context((SynchronizedField, self.setting_descriptor)):
                if self.set_throu_get:
                    value = self.__get__(instance, type(instance))
                self.setting_descriptor.__set__(instance, value)


class DateTimeSpan(TimeSpan):
    '''A *continuous* span of time (with datetime at each boundary).

    `DateTimeSpan`:class: is a minor extension of `TimeSpan`:class:, and is a
    subclass.

    DateTimeSpan objects are iterable.  They yield exactly two datetimes:
    first the start date, and then the end date::

       >>> ts = DateTimeSpan('2017-08-01 11:00', '2017-09-01 23:00')
       >>> tuple(ts)
       (datetime(2017, 8, 1, 11, 0), date(2017, 9, 1, 23, 0))

    The API of DateTimeSpan is just the natural transformation of the API of
    `TimeSpan`:class:.

    The `start_date` and `end_date` attributes are interlocked with the
    `start_datetime` and `end_datetime`.  By changing `start_date`, you also
    change `start_datetime` with the same date at 00:00 without tzinfo.  By
    setting `start_datetime` you also update `start_date`.  By setting
    `end_date` you also update `end_datetime` with the same date at 23:59:59
    without tzinfo.

    .. versionadded:: 1.9.7

    .. warning:: DateTimeSpan is provided on a provisional basis.  Future
       releases can change its API or remove it completely.

    '''
    start_datetime = SynchronizedField(
        DateTimeField('start_datetime', nullable=True),
        TimeSpan.start_date,
    )
    end_datetime = SynchronizedField(
        DateTimeField('end_datetime', nullable=True, prefer_last_minute=True),
        TimeSpan.end_date,
    )
    start_date = SynchronizedField(
        TimeSpan.start_date,
        start_datetime.descriptor,
    )
    end_date = SynchronizedField(
        TimeSpan.end_date,
        end_datetime.descriptor,
    )

    def __init__(self, start_datetime=None, end_datetime=None):
        # Don't call super because our fields are synchronized.
        self.start_datetime = start_datetime
        self.end_datetime = end_datetime

    @classmethod
    def from_datetime(self, dt):
        # type: (datetime) -> DateTimeSpan
        '''Return a new date time span that covers a single `datetime`.

        If `dt` is actually a date, the start_datetime will be at '00:00:00'
        and the end_datetime will be at '23:59:59'.

        '''
        return self(start_datetime=dt, end_datetime=dt)

    @classmethod
    def from_timespan(self, ts):
        # type: (TimeSpan) -> DateTimeSpan
        '''Return a new date time span from a timespan.

        Notice the start datetime will be set at '00:00:00' and the end
        datetime at '23:59:59'.

        If `ts` is already a DateTimeSpan, return it unchanged.

        '''
        if isinstance(ts, DateTimeSpan):
            return ts
        else:
            return self(start_datetime=ts.start_date,
                        end_datetime=ts.end_date)

    @property
    def past_unbound(self):
        # type: () -> bool
        'True if the time span is not bound into the past.'
        return self.start_datetime is None

    @property
    def future_unbound(self):
        # type: () -> bool
        'True if the time span is not bound into the future.'
        return self.end_datetime is None

    @property
    def unbound(self):
        # type: () -> bool
        '''True if the time span is `unbound into the past <past_unbound>`:attr: or
        `unbount into the future <future_unbound>`:attr: or both.

        '''
        return self.future_unbound or self.past_unbound

    @property
    def bound(self):
        # type: () -> bool
        'True if the time span is not `unbound <unbound>`:attr:.'
        return not self.unbound

    @property
    def valid(self):
        # type: () -> bool
        '''A bound time span is valid if it starts before it ends.

        Unbound time spans are always valid.

        '''
        if self.bound:
            return self.start_datetime <= self.end_datetime
        else:
            return True

    def __contains__(self, other):
        # type: (date) -> bool
        '''Test if datetime `other` is in the datetime span.

        If `other` is a `~datetime.date`:class:, we convert it to a naive
        datetime at midnight (00:00:00).

        '''
        if isinstance(other, date):
            if not isinstance(other, datetime):
                other = datetime(  # type: ignore
                    other.year,
                    other.month,
                    other.day
                )
            if self.start_datetime and self.end_datetime:
                return self.start_datetime <= other <= self.end_datetime
            elif self.start_datetime:
                return self.start_datetime <= other
            elif self.end_datetime:
                return other <= self.end_datetime
            else:
                return True
        else:
            return False

    def overlaps(self, other):
        # type: (TimeSpan) -> bool
        '''Test if the time spans overlaps.'''
        return bool(self & other)

    def isdisjoint(self, other):
        # type: (TimeSpan) -> bool
        return not self.overlaps(other)

    def __le__(self, other):
        # type: (TimeSpan) -> bool
        'True if `other` is a superset.'
        return (self & other) == self

    issubset = __le__

    def __lt__(self, other):
        # type: (TimeSpan) -> bool
        'True if `other` is a proper superset.'
        return self != other and self <= other

    def __gt__(self, other):
        # type: (TimeSpan) -> bool
        'True if `other` is a proper subset.'
        return self != other and self >= other

    def __ge__(self, other):
        # type: (TimeSpan) -> bool
        'True if `other` is a subset.'
        # Notice that ge is not the opposite of lt.
        return (self & other) == other

    issuperset = covers = __ge__

    def __iter__(self) -> Iterator[datetime]:  # type: ignore
        yield self.start_datetime
        yield self.end_datetime

    def __getitem__(self, index) -> datetime:  # type: ignore
        this = tuple(self)
        return this[index]

    def __eq__(self, other):
        if isinstance(other, date):
            other = type(self).from_datetime(other)
        elif isinstance(other, TimeSpan) and \
             not isinstance(other, DateTimeSpan):   # noqa
            other = self.from_timespan(other)
        elif not isinstance(other, DateTimeSpan):
            return NotImplemented
        return (self.start_datetime == other.start_datetime and
                self.end_datetime == other.end_datetime)

    def __hash__(self):
        return hash((DateTimeSpan, self.start_datetime, self.end_datetime))

    def __and__(self, other):
        # type: (TimeSpan) -> DateTimeSpan
        '''Get the date time span that is the intersection with another time span.

        If two time spans don't overlap, return the object
        `EmptyTimeSpan`:any:.

        If `other` is not a DateTimeSpan we try to create one.  If `other` is
        a date/datetime, we create use `from_datetime`:meth:.  If `other` is
        TimeSpan we use `from_timespan`:meth:.  Other types are passed
        unchanged to the constructor.

        '''
        from xotl.tools.infinity import Infinity
        if isinstance(other, _EmptyTimeSpan):
            return other
        elif isinstance(other, date):
            other = DateTimeSpan.from_datetime(other)
        elif isinstance(other, TimeSpan):
            other = DateTimeSpan.from_timespan(other)
        elif not isinstance(other, TimeSpan):
            raise TypeError("Invalid type '%s'" % type(other).__name__)
        start = max(
            self.start_datetime or -Infinity,
            other.start_datetime or -Infinity
        )
        end = min(
            self.end_datetime or Infinity,
            other.end_datetime or Infinity
        )
        if start <= end:
            if start is -Infinity:
                start = None
            if end is Infinity:
                end = None
            return type(self)(start, end)
        else:
            return EmptyTimeSpan
    __mul__ = __rmul__ = __rand__ = __and__

    def __bool__(self):
        # type: () -> bool
        return True

    __nonzero__ = __bool__

    def __lshift__(self, delta):
        # type: (Union[int, timedelta]) -> DateTimeSpan
        '''Return the date time span displaced to the past in `delta`.

        :param delta: The number of days to displace.  It can be either an
                      integer or a `datetime.timedelta`:class:.  The integer
                      will be converted to ``timedelta(days=delta)``.

        .. warning:: Python does have a boundaries for the dates it can
           represent, so displacing can cause OverflowError.

        '''
        import numbers
        if isinstance(delta, numbers.Integral):
            delta = timedelta(days=delta)
        start = self.start_datetime - delta if self.start_datetime else None
        end = self.end_datetime - delta if self.end_datetime else None
        return type(self)(start, end)

    def __rshift__(self, delta):
        # type: (Union[int, timedelta]) -> DateTimeSpan
        '''Return the date time span displaced to the future in `delta`.

        :param delta: The number of days to displace.  It can be either an
                      integer or a `datetime.timedelta`:class:.  The integer
                      will be converted to ``timedelta(days=delta)``.

        .. warning:: Python does have a boundaries for the dates it can
           represent, so displacing can cause OverflowError.

        '''
        return self << -delta

    def intersection(self, *others):
        # type: (TimeSpan) -> DateTimeSpan
        'Return ``self [& other1 & ...]``.'
        import operator
        from functools import reduce
        return reduce(operator.mul, others, self)

    def diff(self, other):
        # type: (TimeSpan) -> Tuple[DateTimeSpan, DateTimeSpan]
        '''Return the two datetime spans which (combined) contain all the
        seconds in `self` which are not in `other`.

        Notice this method returns a tuple of exactly two items.

        If `other` and `self` don't overlap, return ``(self, EmptyTimeSpan)``.

        If ``self <= other`` is True, return the tuple with the empty time
        span in both positions.

        Otherwise `self` will have some datetimes which are not in `other`;
        there are possible three cases:

        a) other starts before or at self's start datetime; return the empty
           time span and the datetime span from the second after
           `other.end_datetime` up to `self.end_datetime`

        b) other ends at or after self's end date; return the datetime span
           from `self.start_datetime` up to the second before
           `other.start_datetime` and the empty time span.

        c) `other` is fully contained in `self`; return two non-empty datetime
           spans as in the previous cases.

        '''
        if not self & other:
            return self, EmptyTimeSpan
        other = self & other
        if self == other:
            return EmptyTimeSpan, EmptyTimeSpan
        else:
            assert self > other
            sec = timedelta(seconds=1)
            if self.start_datetime == other.start_datetime:
                return (EmptyTimeSpan,
                        DateTimeSpan(other.end_datetime + sec,
                                     self.end_datetime))
            elif self.end_datetime == other.end_datetime:
                return (DateTimeSpan(self.start_datetime,
                                     other.start_datetime - sec),
                        EmptyTimeSpan)
            else:
                return (DateTimeSpan(self.start_datetime,
                                     other.start_datetime - sec),
                        DateTimeSpan(other.end_datetime + sec,
                                     self.end_datetime))

    def __repr__(self):
        start, end = self
        return 'DateTimeSpan(%r, %r)' % (
            start.isoformat().replace('T', ' ') if start else None,
            end.isoformat().replace('T', ' ') if end else None
        )

    __str__ = __repr__


del IntEnum
